
UNIT FXModule;

{------------------------------------------------------------------------------------
	File:		FXModule.p
	By:		Bill Johnson & Ron Duritsch
	Date:		6.3.92
	Version:	1.0
	Revisions:	

What Are Finder Xtensions?
	Finder Xtensions are modules that can be dropped in a folder named 'Finder Xtensions' in
	the System Folder.  If the user is running the sysem 7 extension 'FX', the name of the
	module will be added to the 'FX' menu in the Finder.  When the user selects that item from 
	the menu, the 'FX' init will build a list of the items that are selected in the Finder and
	this code resource will be loaded and executed to operate on the selected files & folders.
What Is This File?
	This unit is the pascal shell that can be used to create a Finder Xtensions module.  When 
	compiled with mpw using the accompanying 'FXModule.make' make file, the file generated 
	can be dropped into the 'Finder Xtensions' folder and the FX init will do the rest.
What is the format of a Finder Xtension module?
	When compiled, this code will be in a file with a type of 'Xten' and creator of 'Blbd'.  The
	resource type will be 'Xten' and the ID# will be 1968.
When is this code module executed?
	This code resource will be loaded, locked and jumped to from within a faceless background 
	application.  Therefore it cannot put up windows or attempt to communicate with the user
	except through the error reporting done by the Finder Xtension shell (using the 'statusStr'
	field) or the notification manager.  In future versions Xtensions will be able to put up a
	window to get info from the user.
How much memory is available?
	If your module requires more than 5k of heap space, it must be allocated as temporary 
	memory.
 ------------------------------------------------------------------------------------
 }

INTERFACE

{$R+}			{ range checking }

USES
      MemTypes, OSIntf, ToolIntf, SANE, Files;


CONST
	maxSelectedItems	= 256;
	noFurtherProcessing	= 'xxxx';


TYPE
{$SETC BufferVersion2 := TRUE}

	ReplyRecord = record
		finished	: boolean;
		reserved	: boolean;
{$IFC BufferVersion2}
		bufVers	: integer;
{$ENDC}
		errorID		: integer;
		statusStr	: str255;		{ if the module terminates with an error, put explanatory msg here }
		FXShellAction	: OSType;	{ put std apple event token here if you want to specify further processing }
{$IFC BufferVersion2}
		aeTarget	: OSType;
{$ENDC}
		end;
						
	SelectedItemRecord = record
		IsFolder		: boolean;
		IsApplication	: boolean;
		IsAliasFile	: boolean;
		FinderFlags	: integer;
		PhysicalSize	: longint;
		LogicalSize	: longint;
		CreatedDate	: longint;
		ModifiedDate	: longint;
		parentVRef	: integer;
		parentDirID	: longint;
		nameString	: Str31;
		CASE integer OF
			1: (			{ folder }
				fldrDirID		: longint;
		{		fldrVRef		: integer;		}
			);
			2: (			{ file }
				fileID			: longint;
				fileType		: OSType;
				fileCreator	: OSType;
			);
		end;

	SelectionBufferRecord = record
		status			: OSType;	{ may hold values 'idle', 'data', 'busy', or 'done'. }
		moduleName	: str31;
{$IFC BufferVersion2}
		evtModifiers	: integer;
		reserved1	: longint;
		reserved2	: longint;
		privateData	: array [1..8] of longint;	{ 32 bytes of data if module is going to get called multiple times }
{$ELSEC}
		privateData	: array [1..32] of longint;	{ 128 bytes of data if module is going to get called multiple times }
{$ENDC}
		resultInfo		: ReplyRecord;
		numFiles		: integer;
		numFolders	: integer;
		selection		: array [1..maxSelectedItems] of SelectedItemRecord;
		end;
	SelectionBufferPtr = ^SelectionBufferRecord;


	PROCEDURE FinderModule (selectionPtr: SelectionBufferPtr);


IMPLEMENTATION


	PROCEDURE DeleteArrayItem (startItem, NumItems, ItemSize: integer; 
												arrayHeadr, blockSize: longint; arrayPtr: univ Ptr);
	FORWARD;

{------------------------------------------------------------------------------------
How the SelectionBuffer is used.
	The selection buffer is a non-relocatable block of memory used to pass all data that is 
	used to implement Finder Xtensions.  The buffer can be in one of 4 states.  The current
	state is indicated by the contents of the field 'status' at the head of the buffer.  
	During execution of an FX module, the status field will always contain the value 'busy'.
		'idle'	 set initially by background app
			no data in buffer.  no operation in progress
		'data'	 set by init when user chooses a menu item & data is in buffer
			A list of selected items has been placed in the buffer by the FX shell, but has not
			been operated on.
		'busy'	 set by background app just before jumping to code module
			The selected Finder Xtension module is in the process of operating on the data in
			the buffer.  If the 'finished' flag in the 'resultInfo' record is FALSE when the module
			returns control to the FX shell, then the shell will continue to call the module until
			the 'finished' flag is set to TRUE.  This allows the Finder Xtension module to do
			lengthy operations while giving up time to other processes.  The 'status' field will
			contain the value 'busy' until the module is finished.  Any data that is needed across
			calls can be put in the field 'privateData'.
		'done'	 set by background app when module terminates.  Reset to 'idle' by init after error, if any, is reported
			Indicates that a module is terminating.  If an error is indicated it will be reported
			to the user in the Finder.
 ------------------------------------------------------------------------------------}


PROCEDURE FinderModule (selectionPtr: SelectionBufferPtr);
{  This routine must come first in the unit.  Any routines that it calls need to be declared FORWARD }
CONST
	HeaderSize	= sizeof (SelectionBufferRecord) - (maxSelectedItems * sizeOf (SelectedItemRecord));
VAR
	index, ErrID	: integer;
	itemTotal	: integer;
BEGIN
	ErrID := noErr;
	with selectionPtr^ do
		begin
			itemTotal := numFiles + numFolders;
			for index := itemTotal downto 1 do
				begin
					IF (selection[index].IsFolder)
					THEN
						begin	{ process folder }
							;
						end
					ELSE
						begin	{ process file }
							sysbeep (1);
							DeleteArrayItem (index, 1, sizeof (SelectedItemRecord), HeaderSize, 
																	sizeof(SelectionBufferRecord), selectionPtr);
							numFiles := numFiles - 1;
						end;
				end;

			with resultInfo do
				begin
					finished := TRUE;
{$IFC BufferVersion2}
					bufVers := 1;
{$ENDC}
					IF (ErrID = noErr)
					THEN
						begin
							errorID := noErr;
		{ put Finder apple event token here if you want to specify an action for the Finder to perform on the items in the selection list }
							FXShellAction := 'sope';		{ open the selection }
{$IFC BufferVersion2}
							aeTarget := 'MACS';
{$ENDC}
						end
					ELSE
						begin
							errorID := ErrID;
							statusStr := 'This is what went wrong...';	
							FXShellAction := noFurtherProcessing;
						end;
				end;
		end;
END;


PROCEDURE DeleteArrayItem (startItem, NumItems, ItemSize: integer; 
												arrayHeadr, blockSize: longint; 
												arrayPtr: univ Ptr);
{	assume a _one_ based array. You know, array[1..n] of integer, not array[0..n-1] of integer...	}
		var
			totlNum: integer;
			lastDelete: integer;
			moveSize: longint;
			destPtr: Ptr;
			sourcePtr: Ptr;
			HdlFlags: SignedByte;
	begin
		totlNum := (blockSize - arrayHeadr) div ItemSize;
		lastDelete := startItem - 1 + NumItems;

		if (lastDelete < totlNum) then
			begin
				destPtr := pointer(ord(arrayPtr) + ((startItem - 1) * ItemSize) + arrayHeadr);
				sourcePtr := pointer(ord(destPtr) + (NumItems * ItemSize));
				moveSize := (totlNum - lastDelete) * ItemSize;
				BlockMove(sourcePtr, destPtr, moveSize);
			end;
	end;


END.

