
UNIT EditTextCDEV;


INTERFACE


USES
	MemTypes, QuickDraw, OSIntf, ToolIntf, PackIntf, GestaltEqu;


CONST
		textItm		= 1;			{first editTExt item in cdev}
		maxSelectedItems	= 256;

		kIdleStatus			= 'idle';
		kDataReadyMsg		= 'data';
		kCodeBusyMsg		= 'busy';
		kCodeDoneMsg		= 'done';
		kLaunchCdevMsg	= 'cdev';
		kLaunchIntfMsg		= 'blbd';
		kCdevBusyMsg		= 'cpip';

		kGestaltSignature			= 'FXtn';


TYPE
{$SETC BufferVersion2 := TRUE}

	ReplyRecord = record
		finished	: boolean;
		reserved	: boolean;
{$IFC BufferVersion2}
		bufVers	: integer;
{$ENDC}
		errorID		: integer;
		statusStr	: str255;		{ if the module terminates with an error, put explanatory msg here }
		FXShellAction	: OSType;	{ put std apple event token here if you want to specify further processing }
{$IFC BufferVersion2}
		aeTarget	: OSType;
{$ENDC}
		end;
						
	SelectedItemRecord = record
		IsFolder		: boolean;
		IsApplication	: boolean;
		IsAliasFile	: boolean;
		FinderFlags	: integer;
		PhysicalSize	: longint;
		LogicalSize	: longint;
		CreatedDate	: longint;
		ModifiedDate	: longint;
		parentVRef	: integer;
		parentDirID	: longint;
		nameString	: Str31;
		CASE integer OF
			1: (			{ folder }
				fldrDirID		: longint;
		{		fldrVRef		: integer;		}
			);
			2: (			{ file }
				fileID			: longint;
				fileType		: OSType;
				fileCreator	: OSType;
			);
		end;

	SelectionBufferRecord = record
		status			: OSType;	{ may hold values 'idle', 'data', 'busy', or 'done'. }
		moduleName	: str31;
{$IFC BufferVersion2}
		evtModifiers	: integer;
		reserved1	: longint;
		reserved2	: longint;
		privateData	: array [1..8] of longint;	{ 32 bytes of data if module is going to get called multiple times }
{$ELSEC}
		privateData	: array [1..32] of longint;	{ 128 bytes of data if module is going to get called multiple times }
{$ENDC}
		resultInfo		: ReplyRecord;
		numFiles		: integer;
		numFolders	: integer;
		selection		: array [1..maxSelectedItems] of SelectedItemRecord;
		end;
	SelectionBufferPtr = ^SelectionBufferRecord;

	CDEVRec	= RECORD
			DataPtr	: SelectionBufferPtr;
		END;
	CDEVPtr	= ^CDEVRec;
	CDEVHnd	= ^CDEVPtr;
	


FUNCTION TextCDEV (message, item, numItems, CPanelID: INTEGER;
											VAR theEvent: EventRecord;
											cdevStorage: Handle;
											CPDialog: DialogPtr) : Handle;
												
IMPLEMENTATION


PROCEDURE DoEditCommand (message: INTEGER; CPDialog: DialogPtr);						FORWARD;
FUNCTION GetGestaltPtrReference (VAR DataPtr: SelectionBufferPtr): OSErr;			FORWARD;


FUNCTION TextCDEV (message, item, numItems, CPanelID: INTEGER;
											VAR theEvent: EventRecord;
											cdevStorage: Handle;
											CPDialog: DialogPtr) : Handle;
VAR
	tempChar		: char;
	tempDataPtr	: SelectionBufferPtr;
BEGIN
	IF (message = macDev)
	THEN
		begin
			IF (GetGestaltPtrReference (tempDataPtr) = noErr)
			THEN
				begin
					IF (tempDataPtr^.status = kLaunchIntfMsg)
					THEN
						begin
							tempDataPtr^.status := kCdevBusyMsg;
							TextCDEV := Handle (1);
						end
					ELSE
						begin
							sysbeep (1);		{ another module or cdev is using the buffer }
							TextCDEV := Handle (0);
						end;
				end
			ELSE
				begin
					sysbeep (1);		{ background app has not setup buffer & registered with Gestalt }
					TextCDEV := Handle (0);
				end;
		end
	ELSE
		begin
			IF (cdevStorage <> NIL)
			THEN
				BEGIN
					CASE message OF
						initDev:											{initialize cdev}
							BEGIN
								cdevStorage := NewHandle (SIZEOF(CDEVRec));		{create private storage}
								SelIText (CPDialog, numItems + textItm, 0, 999);	{make caret show up}
								IF (GetGestaltPtrReference (tempDataPtr) = noErr)
								THEN
									CDEVHnd (cdevStorage)^^.DataPtr := tempDataPtr
								ELSE
									CDEVHnd (cdevStorage)^^.DataPtr := nil;
							END;
						hitDev:;
						closeDev:
							begin
								with CDEVHnd (cdevStorage)^^.DataPtr^.resultInfo do
									begin
										finished := TRUE;
										errorID := noErr;
										statusStr := '';
										FXShellAction := 'sope';
{$IFC BufferVersion2}
										bufVers := 1;
										aeTarget := 'MACS';
{$ENDC}
									end;
							end;
						nulDev:;
						updateDev:;
						activDev:;
						deActivDev:;
						keyEvtDev:											{respond to key down}
							BEGIN
								{first, get the character}
								tempChar := CHR(BAnd(theEvent.message, charCodeMask));
								{then see if the command key was down}
								IF (BAnd(theEvent.modifiers, cmdKey) <> 0)
								THEN
									BEGIN
										message := nulDev;							{start off with no message}
										theEvent.what := nullEvent;					{wipe out event}
										CASE tempChar OF							{set appropriate message}
											'X','x':
												message := cutDev;
											'C','c':
												message := copyDev;
											'V','v':
												message := pasteDev;
											END;
										DoEditCommand(message, CPDialog);			{let edit command handler take it}
									END;
							END;
						undoDev:;
						cutDev, copyDev, pasteDev, clearDev:
							DoEditCommand(message, CPDialog);				{respond to edit command}
						END;  {CASE message}
					TextCDEV := cdevStorage;
					{if cdevStorage = NIL then ControlPanel will put up memory error}
				END;
		END;  {cdevStorage <> NIL}
END; {TextCDEV}


PROCEDURE DoEditCommand (message: INTEGER; CPDialog: DialogPtr);
BEGIN
	CASE message OF
		cutDev:
			DlgCut (CPDialog);
		copyDev:
			DlgCopy (CPDialog);
		pasteDev:
			DlgPaste (CPDialog);
		clearDev:
			DlgDelete (CPDialog);
		END;
END; {DoEditCommand}


FUNCTION GetGestaltPtrReference (VAR DataPtr: SelectionBufferPtr): OSErr;		{ return error }
VAR
	gestaltErr		: OSErr;
	memLocation	: longint;
BEGIN
	DataPtr := nil;
	GetGestaltPtrReference := -1;
{	IF (GestaltAvailable)
	THEN
		begin	}
			gestaltErr := Gestalt (kGestaltSignature, memLocation);
			GetGestaltPtrReference := gestaltErr;
			IF (gestaltErr = noErr)
			THEN
				DataPtr := SelectionBufferPtr (memLocation);
	{	end;	}
END;


END.  {EditTextCDEV}
